﻿//
// Code very loosely based on a script by loic_aigon, posted on the InDesign scripting forums.
// Enhanced, bugfixed and cleaned up by Peter Werner (PeterWerner@gmx.ch).
//
// TODO:
//
//  - Instead of removing the first frame, why not set its column count to 1 and
//    use it as the first one of the new text frames? That also saves us having
//    to copy the contents and would make the script slightly more efficient.
//
// KNOWN ISSUES/FIXMES:
//
//  - Undo undoes in steps. I don't know if InDesign exposes
//    a transaction-based undo framework to scripts, but I guess not.
//    But using the C++ API is most likely not worth the extra effort.
//  - Transforms are not yet handled correctly
//  - There is some weird problems when loading this as a startup script.
//    This is probably related to us using the session engine.
//  - If identifiers like locLocalize() are used in any other script running
//    in the session engine, the later ones overwrite stuff from the earlier
//    ones, which can lead to an earlier script calling the wrong version.
//    I think I heard we can define our own engines with persistence support
//    by specifying #targetengine "somethingArbitrary", but we'd have to
//    check on this.
//
// FIXES/ENHANCEMENTS OVER ORIGINAL SCRIPT:
//
//  - Script Labels are no longer used to identify the frames. That way 
//    existing script labels are not lost, and the command can also be
//    run more than once per document without clearing all script labels first.
//  - Main code of the command has been simplified, there was quite a bit of
//    unnecessary confusion going on with commands inside a loop that only
//    needed to be executed once and so on
//  - The script now installs itself in InDesigns Object menu as 
//    "Object > Break Into Columns". The item even gets enabled/disabled
//    depending on whether a text (frame) is currently selected or not
//  - The script now handles frames with only one column correctly
//  - Z order is preserved correctly (should be checked again)
//  - Frame attributes such as stroke, fill etc. are preserved
//  - Locked frames are handled correctly (i.e. the command is greyed out)
//  - Text threads are dealt with correctly
//  - Assigned InCopy frames are handled correctly. The original link to the
//    story file is preserved instead of the content showing up 
//    as "Unassigned InCopy Content" in the Assignments panel
//  - Assigned InCopy frames no longer need to be checked out in order for
//    the script to work (the original one used to throw an error saying the
//    frame was locked I believe)
//  - The new frames now end up on the same layer as the original one
//  - The new frames now end up on the same page/spread as the original one
//    (previously they were created on the first spread of the document)
//
//=============================================================================
#target "InDesign"        // Run this script in InDesign
#targetengine "session"   // We need persistence since we install event handlers
//=============================================================================
// Localized string constants. Do not use directly, use locLocalize() instead
var _localizedStrings = Array();
//-----------------------------------------------------------------------------
// This function initializes the localization table with strings in the
// current language.
//
// Once you have called this function, you can use locLocalize() to 
// get a localized string for the given key.
//
// To define your own resource strings, edit the _locInitializeXXX()
// functions.
//
// Example (assumes that you defined a key string "$TEST/MyMessage"):
//
//     locInitialize();
//
//     alert (
//         locLocalize( "$TEST/MyMessage" )
//     );
//
//
function locInitialize( )
{
	switch( $.locale )
	{
		case( "fr_FR" ): _locInitializeFr(); return;
		case( "de_DE" ): _locInitializeDe(); return;
		case( "it_IT" ): _locInitializeIt(); return;
		default: 	     _locInitializeEn(); return;
	}
}
//-----------------------------------------------------------------------------
function _locInitializeFr()
{
	_localizedStrings[ "$BC/ScriptAlreadyLoadedMsg" ] = "Scripte déjà chargé. Cherchez dans le menu objet :)";
	_localizedStrings[ "$BC/BreakColumnsMenuItem"   ] = "Séparer colonnes"
	_localizedStrings[ "$BC/AssertionFailed"        ] = "Erreur:";
}
//-----------------------------------------------------------------------------
function _locInitializeDe()
{
	_localizedStrings[ "$BC/ScriptAlreadyLoadedMsg" ] = "Script bereits geladen. Im Objekt-Menü sollte ein neuer Eintrag aufgetaucht sein :)";
	_localizedStrings[ "$BC/BreakColumnsMenuItem"   ] = "Spalten in Rahmen konvertieren";
	_localizedStrings[ "$BC/AssertionFailed"        ] = "Fehler:";
}
//-----------------------------------------------------------------------------
function _locInitializeIt()
{
	_localizedStrings[ "$BC/ScriptAlreadyLoadedMsg" ] = "Script è già carico :)";
	_localizedStrings[ "$BC/BreakColumnsMenuItem"   ] = "Dividere colonne";
	_localizedStrings[ "$BC/AssertionFailed"        ] = "Errore:";
}
//-----------------------------------------------------------------------------
function _locInitializeEn()
{
	_localizedStrings[ "$BC/ScriptAlreadyLoadedMsg" ] = "Script already loaded. Check your Object menu :)";
	_localizedStrings[ "$BC/BreakColumnsMenuItem"   ] = "Break Into Columns";
	_localizedStrings[ "$BC/AssertionFailed"        ] = "Error:";
}
//-----------------------------------------------------------------------------
// Use this to get a localized version of the string identified by the given
// key string (key strings look like "$BC/SomeString").
function locLocalize( keyString )
{
	var theString = _localizedStrings[ keyString ];
	
	if( theString == null )
		throw "Invalid key string for localization (" + keyString ")" ;
	
	return theString;
}
//=============================================================================
//
// Internal debugging function to ensure that some assumptions made
// when writing the code always hold true.
//
function dbgAssert( value, description )
{
	if( !value )
		throw locLocalize( "$BC/AssertionFailed:" ) + " " + description;
}
//-----------------------------------------------------------------------------
//
// This function returns a text frame for the current selection, or null if
// no text frame or text in a frame is currently selected.
//
function getSelectedTextFrame()
{
	if( app.selection[0] == null )
		return null;
		
	if( app.selection[0].parent.constructor.name == "Story" ) 
		return app.selection[0].parentTextFrames;

	if( app.selection[0].constructor.name == "TextFrame" )
		return app.selection[0];

	return null;
}
//-----------------------------------------------------------------------------
//
// This function breaks an existing multi-column text frame into individual frames,
// one for each column.
//
function breakColumns()
{
	// Make sure the selection is right. It should be when the function is called
	// since it is the responsibility of the calling function to ensure that.
	dbgAssert (
		isValidContext(),
		"isValidContext() == false"
	);
		
	// Get the current text frame. Should never fail since we validated the context before.
	// An error here is most likely caused by a bug in isValidContext().
	var tf = getSelectedTextFrame();		
	dbgAssert( tf != null, "getSelectedTextFrame() returned null" );
		
	var clmwidth  = Math.round( tf.textFramePreferences.textColumnFixedWidth * 100 ) / 100;
	var spclm     = Math.round( tf.textFramePreferences.textColumnGutter     * 100 ) / 100;
	var clmnb     = tf.textColumns.length;
	var tfpos     = tf.geometricBounds;
	var tfwh      = tf.geometricBounds[3] - tf.geometricBounds[1];
	var contenttf = tf.contents;
	
	var prev_frame  = tf;
	
	for( i = 0; i < clmnb; i++ )
	{
		// Calculate the bounds for the new text frame
		var y1 = tfpos[ 0 ];
		var x1 = tfpos[ 1 ] + i * ( clmwidth + spclm );
		var y2 = tfpos[ 2 ];
		var x2 = x1 + clmwidth;
		
		// Make a new text frame with identical attributes
		var ntf = tf.duplicate();
		
		// Remove the content
		ntf.contents = "";
		
		// Set the column count to one
		ntf.textFramePreferences.textColumnCount = 1;
		
		// Size and position it like the original column
		ntf.geometricBounds = [ y1, x1, y2, x2 ];	
					
		// Link the frame to the previous one. The first one
		// is linked to the original frame, thus maintaining
		// correct links to an InCopy story if the frame was
		// an InCopy assigned frame.
		prev_frame.nextTextFrame = ntf;		

		// This is going to be threaded to the next frame
		// we are going to generate
		prev_frame = ntf;
	}

	// Delete the original text frame
	tf.remove();
}
//-----------------------------------------------------------------------------
//
// This function returns true if the current application state permits
// running the action, i.e. we need a document and a selected text frame
// with more than one column etc.
//
function isValidContext()
{
	// We need an active document
	if( app.documents.length == 0 )
		return false;
		
	// We need a selection for this to work
	if( app.selection.length == 0 )
		return false;
	
	// We can only operate on a text frame
	if( app.selection[0].parent.constructor.name != "Story" && app.selection[0].constructor.name != "TextFrame" )
		return false;

	var tf = getSelectedTextFrame();
	
	// We need a valid frame selection
	if( tf == null )
		return false;

	// We can't work on a locked frame. We could theoretically unlock
	// and then relock the frame, but that would kind of defeat the purpose
	// of letting the user lock a frame. The command being greyed out 
	// is more like what a user would expect, even though InDesign lets 
	// users change the number of columns on a locked frame.
	if( tf.locked == true )
		return false;

	// We need more than one column.
	if( tf.textColumns.length < 2 )
		return false;
		
	// Everything's fine
	return true;
}
//-----------------------------------------------------------------------------
//
// This is called when the parent menu is popped up. Here we enable/disable
// our menu item depending on the current selection type.
//
function beforeDisplayBreakColumns()
{
	// Get a reference to the menu command we are working on
	var action = app.scriptMenuActions.item( locLocalize( "$BC/BreakColumnsMenuItem" ) );

	// Enable/disable the menu command
	action.enabled = isValidContext();
}
//-----------------------------------------------------------------------------
//
// This function returns a reference to any of the top level application
// menus, such as "File", "Layout", or "Window". You can use the regular
// English menu names, they are automatically localized by the function.
//
function getTopLevelMenu( name )
{
	// Get a reference to the main menu bar
	var mainMenu = app.menus.item( "Main" );
	
	// Translate the name of the menu. Main menu items have the
	// resource identifier "$ID/MenuItemCaptionInEnglishVersion",
	// so this should work just fine.
	var localizedName = app.translateKeyString( "$ID/" + name );
	
	// Find the menu
	var theMenu = mainMenu.menuElements.item( localizedName );
	
	// Make sure it worked. It should if the argument
	// was valid.
	dbgAssert ( 
		theMenu != null,
		"No such menu (" + name + ")"
	);
		
	return theMenu;
}
//-----------------------------------------------------------------------------
//
// Installs a new command into InDesign's menu system
//
// menu                 : The menu to add the item to. For example, use
//                        app.menus.items( "Main" ) for the top-level menu, 
//                        or you can use any sub menu as well of course.
// title                : Caption of the new menu item
// callback             : The function to be called when the command is invoked
// beforeDisplayCallback: Function to call before the parent menu pops up.
//                        Use it to enable/disable the item based on context/selection.
//                        May be null if not needed.
//
function installMenu( menu, title, invokeCallback, beforeDisplayCallback )
{
	// Create an action and attach an event listener
	var hAction      = app.scriptMenuActions.add( title );
	var hEvtListener = hAction.eventListeners.add( "onInvoke", invokeCallback );
	
	if( beforeDisplayCallback != null )
		hAction.eventListeners.add( "beforeDisplay", beforeDisplayCallback );
	
	// Add the action to the menu
	menu.menuItems.add( hAction );

	// FIXME: Do we get a memory leak if we don't store hAction/hEvtListener
	//        and then somehow unregister it at some point? Is there even
	//        such a thing as a memory leak in JavaScript?
}
//-----------------------------------------------------------------------------
//
// The scripts mainline. It basically just installs a menu item in InDesign's
// Object menu.
//
function main()
{
	// Initialize the localization table
	locInitialize( );
	
	// Prevent the script from installing more than one menu item
	// it it is run more than once.
	// You should disable the "return" call in debug versions to
	// allow reloading of the script.
	// FIXME: This somehow causes trouble if the script is running as
	// a startup script, don't know why. 
	if( app.scriptMenuActions.item( locLocalize( "$BC/BreakColumnsMenuItem" ) ) != null )
	{
		alert ( 
			locLocalize( "$BC/ScriptAlreadyLoadedMsg" )
		);
		
		return;
	}	
	
	// Get a reference to InDesign's Object menu
	var objectMenu = getTopLevelMenu( "Object" );
	
	// Add a separator to the object menu
	objectMenu.menuSeparators.add();
	
	// Add an item to the Object menu
	installMenu (
		objectMenu,
		locLocalize( "$BC/BreakColumnsMenuItem" ),
		breakColumns,
		beforeDisplayBreakColumns
	);
}
//=============================================================================
// Invoke main (JavaScript sucks...)
main();
//=============================================================================